/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono;

import com.qitsoft.qimono.runners.QimonoJUnitRunner;
import com.qitsoft.qimono.stubbing.FieldStubbing;
import com.qitsoft.qimono.stubbing.ReflectiveOngoingStubbing;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import org.mockito.Mockito;
import org.mockito.exceptions.misusing.MissingMethodInvocationException;
import org.mockito.stubbing.OngoingStubbing;

/**
 * The main class used to mock methods. This class could be used in place of
 * {@link org.mockito.Mockito} because it extends from it and all of the methods
 * are available. Also it adds some additional methods allowing stubbing the
 * final, private,protected,package and static methods. Also it allows setting and 
 * getting the values of private/protected/package fields even if they do not have
 * setters or getters.
 * 
 * Base usages which differs from Mockito usages.
 * 
 * <ol>
 *  <li><a href="#non-public-methods">Stubbing inaccessible method due private/protected visibility.</a></li>
 *  <li><a href="#static-methods">Stubbing static methods.</a></li>
 *  <li><a href="#final-methods">Stubbing final methods.</a></li>
 *  <li><a href="#final-classes">Stubbing methods in final class.</a></li>
 *  <li><a href="#non-public-fields">Setting and gettinf the value of inaccessible fields due private/protected visibility.</a></li>
 * </ol>
 * 
 * <a name="non-public-methods" ></a>
 * <h3>Stubbing inaccessible method due private/protected visibility.</h3>
 * 
 * You can stub private methods by invoking the construction like this:
 * <pre><code>
 * QiMockito.when().call("privateMethods", String.class).from(mockObj).withArgs("hello").thenReturn("the mock result");
 * </code></pre>
 * 
 * If your method does not have any arguments you can use:
 * <pre><code>
 * QiMockito.when().call("privateMethods").from(mockObj).thenReturn("the mock result");
 * </code></pre>
 * 
 * Also you can avoid using {@link ReflectiveOngoingStubbing#from(java.lang.Object) }
 * by passing the mock object in {@link #when(java.lang.Object, java.lang.String) }. 
 * This method also avoid using {@link ReflectiveOngoingStubbing#call(java.lang.String, java.lang.Class[]) }
 * for methods without arguments.
 * 
 * If your method does not return (is void method) you can use the same pattern, so
 * there is no need to use {@link org.mockito.Mockito#doReturn(java.lang.Object) } etc.
 * 
 * <a name="static-methods" ></a>
 * <h3>Stubbing static methods.</h3>
 * 
 * To stub static methods you should use the following syntax:
 * <pre><code>
 * QiMockito.when().call("privateMethods").from(mockClass.class).thenReturn("the mock result");
 * </code></pre>
 * 
 * If the method have arguments then you should use:
 * <pre><code>
 * QiMockito.when().call("privateMethods", String.class).from(mockClass.class).withArgs("arg").thenReturn("the mock result");
 * </code></pre>
 * 
 * <a name="final-methods" ></a>
 * <h3>Stubbing final methods.</h3>
 * 
 * To stub final methods use the same syntax as Mockito provides as it is not a final
 * method or syntax described above.
 * 
 * <a name="final-classes" ></a>
 * <h3>Stubbing final classes.</h3>
 * 
 * To stub methods from final classes use the same syntax as Mockito provides or 
 * described above. The Qimono transforms the final classes into non final ones.
 * 
 * <a name="non-public-fields" ></a>
 * <h3>Setting and getting the value of inaccessible fields due private/protected visibility.</h3>
 * 
 * To get the private/protected field you should use the following syntax:
 * <pre><code>
 * QiMockito.field("privateField").from(anObject).get();
 * // or
 * QiMockito.field(anyObject, "privateField").get();
 * </code></pre>
 * 
 * You can operate with private fields not only with mocked objects but with any object.
 * To set the value to the field you can use:
 * <pre><code>
 * QiMockito.field("privateField").from(anObject).set(value);
 * // or
 * QiMockito.field(anyObject, "privateField").set(value);
 * </code></pre>
 * 
 * To use QiMockito you have to use the junit runner {@link QimonoJUnitRunner}.
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public class QiMockito extends Mockito {

    public static <T> ReflectiveOngoingStubbing<T> when() {
        return new ReflectiveOngoingStubbing<T>();
    }

    public static OngoingStubbing when(Object object, String methodName) {
        try {
            Method method = object.getClass().getMethod(methodName);
            Object methodResult = method.invoke(object);
            return Mockito.when(methodResult);

        } catch (IllegalAccessException ex) {
            throw new RuntimeException(ex);
        } catch (IllegalArgumentException ex) {
            throw new RuntimeException(ex);
        } catch (InvocationTargetException ex) {
            throw new RuntimeException(ex);
        } catch (NoSuchMethodException ex) {
            throw new MissingMethodInvocationException(ex.getMessage());
        } catch (SecurityException ex) {
            throw new RuntimeException(ex);
        } catch (NullPointerException ex) {
            throw new RuntimeException(ex);
        }
    }

    public static <T> FieldStubbing<T> field(Object object, String fieldName) {
        return new FieldStubbing<T>(object, fieldName);
    }

    public static <T> FieldStubbing<T> field(String fieldName) {
        return new FieldStubbing<T>(fieldName);
    }
    
}
