/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono;

import com.qitsoft.qimono.internal.substitutors.DateSubstitutor;
import java.util.TimeZone;

/**
 * The stubbing object used to make substitutions. To substitute the date and time 
 * you can perform substituting each field separately.
 * After substitution was performed the timer goes on, so if you substitute the
 * time and the call to obtain the date and time is performed after 5 seconds
 * the obtained time will be 5 seconds after the time you set.
 * 
 * To substitute the date you have to use:
 * <pre><code>
 * Substitute.datetime(2010, 02, 03, 04, 05, 06); // To set 3rd of February 04:05:06
 * </code></pre>
 * 
 * And in the testable code when you use {@code new Date()} or other instructions to
 * get the current date and time you will obtain the set date and time before.
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public class Substitute {

    /**
     * Substitute the date and time. The time zone is ignored.
     * 
     * @param year the year.
     * @param month the month of the year. January is 1, February is 2 etc.
     * @param date the day of month.
     * @param hour the hour of the day in 24 hours format.
     * @param minute the minutes.
     * @param second the seconds.
     * @param millisecond the milliseconds.
     */  
    public static void datetime(int year, int month, int date, int hour, int minute, int second, int millisecond) {
        DateSubstitutor.setDateTime(year, month, date, hour, minute, second, millisecond);
    }
    
    /**
     * Substitute the date and time with time zone.
     * 
     * @param year the year.
     * @param month the month of the year. January is 1, February is 2 etc.
     * @param date the day of month.
     * @param hour the hour of the day in 24 hours format.
     * @param minute the minutes.
     * @param second the seconds.
     * @param millisecond the milliseconds.
     * @param timezone the time zone of the setting time.
     */
    public static void datetime(int year, int month, int date, int hour, int minute, int second, int millisecond, TimeZone timezone) {
        DateSubstitutor.setDateTime(year, month, date, hour, minute, second, millisecond, timezone);
    }
    
    /**
     * Substitute the date and time. The milliseconds and time zone is ignored.
     * 
     * @param year the year.
     * @param month the month of the year. January is 1, February is 2 etc.
     * @param date the day of month.
     * @param hour the hour of the day in 24 hours format.
     * @param minute the minutes.
     * @param second the seconds.
     */
    public static void datetime(int year, int month, int date, int hour, int minute, int second) {
        DateSubstitutor.setDateTime(year, month, date, hour, minute, second);
    }
    
    /**
     * Substitute the date and time with time zone. Milliseconds are ignored.
     * 
     * @param year the year.
     * @param month the month of the year. January is 1, February is 2 etc.
     * @param date the day of month.
     * @param hour the hour of the day in 24 hours format.
     * @param minute the minutes.
     * @param second the seconds.
     * @param timezone the time zone of the setting time.
     */
    public static void datetime(int year, int month, int date, int hour, int minute, int second, TimeZone timezone) {
        DateSubstitutor.setDateTime(year, month, date, hour, minute, second, timezone);
    }
    
    /**
     * Substitute the date. The time is ignored.
     * 
     * @param year the year.
     * @param month the month of the year. January is 1, February is 2 etc.
     * @param date the day of month.
     */
    public static void date(int year, int month, int date) {
        DateSubstitutor.setDate(year, month, date);
    }

    /**
     * Substitute the time. Time zone is ignored.
     * 
     * @param hour the hour of the day in 24 hours format.
     * @param minute the minutes.
     * @param second the seconds.
     * @param millisecond the milliseconds.
     */
    public static void time(int hour, int minute, int second, int millisecond) {
        DateSubstitutor.setTime(hour, minute, second, millisecond);
    }

    /**
     * Substitute the time with time zone.
     * 
     * @param hour the hour of the day in 24 hours format.
     * @param minute the minutes.
     * @param second the seconds.
     * @param millisecond the milliseconds.
     * @param timezone the time zone of the setting time.
     */
    public static void time(int hour, int minute, int second, int millisecond, TimeZone timezone) {
        DateSubstitutor.setTime(hour, minute, second, millisecond, timezone);
    }

    /**
     * Substitute the time. Milliseconds and time zone are ignored.
     * 
     * @param hour the hour of the day in 24 hours format.
     * @param minute the minutes.
     * @param second the seconds.
     */
    public static void time(int hour, int minute, int second) {
        DateSubstitutor.setTime(hour, minute, second);
    }

    /**
     * Substitute time. Millisecond field is ignored.
     * 
     * @param hour the hour of the day in 24 hours format.
     * @param minute the minutes.
     * @param second the seconds.
     * @param timezone the time zone of the setting time.
     */
    public static void time(int hour, int minute, int second, TimeZone timezone) {
        DateSubstitutor.setTime(hour, minute, second, timezone);
    }

    /**
     * Substitute the year. Other fields are untouched.
     * 
     * @param year the year.
     */
    public static void year(int year) {
        DateSubstitutor.setYear(year);
    }

    /**
     * Substitute the month. Other fields are untouched.
     * 
     * @param month the month of the year. January is 1, February is 2 etc.
     */
    public static void month(int month) {
        DateSubstitutor.setMonth(month);
    }
    
    /**
     * Substitute the day of the month. Other fields are untouched.
     * 
     * @param date the day of month.
     */
    public static void date(int date) {
        DateSubstitutor.setDate(date);
    }
    
    /** 
     * Substitute the hour. Other fields are untouched.
     * 
     * @param hour the hour of the day in 24 hours format.
     */
    public static void hour(int hour) {
        DateSubstitutor.setHour(hour);
    }
    
    /**
     * Substitute the minute. Other fields are untouched.
     * 
     * @param minute the minutes.
     */
    public static void minute(int minute) {
        DateSubstitutor.setMinute(minute);
    }
    
    /**
     * Substitute the second. Other fields are untouched.
     * 
     * @param second the seconds.
     */
    public static void second(int second) {
        DateSubstitutor.setSecond(second);
    }
    
    /**
     * Substitute the millisecond. Other fields are untouched.
     * 
     * @param millisecond the milliseconds.
     */
    public static void millisecond(int millisecond) {
        DateSubstitutor.setMillisecond(millisecond);
    }
    
    /**
     * Substitute the time zone. Other fields are untouched.
     * 
     * @param timezone the time zone of the setting time.
     */
    public static void timezone(TimeZone timezone) {
        DateSubstitutor.setTimezone(timezone);
    }

}
