/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono.internal;

import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;

/**
 * The utility class used for class transformations. This class has the
 * methods to change the access flags of the class/method/field.
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public class ClassModifierUtils {

    /**
     * Changes the access flags of the class/method/field to be public. If the 
     * class/method/field is private or protected it transforms the access flag 
     * to be public. If the visibility is package (default) it also makes it public.
     * The rest of options it does not touch, eg. if the access field has the
     * option that the member is static it left it untouched.
     * 
     * @param access the access flag to be transformed into public.
     * @return the modified access field made public.
     * 
     * @see org.objectweb.asm.ClassVisitor#visit(int, int, java.lang.String, java.lang.String, java.lang.String, java.lang.String[]) 
     * @see org.objectweb.asm.ClassVisitor#visitMethod(int, java.lang.String, java.lang.String, java.lang.String, java.lang.String[])  
     * @see org.objectweb.asm.ClassVisitor#visitField(int, java.lang.String, java.lang.String, java.lang.String, java.lang.Object) 
     * @see org.objectweb.asm.Opcodes#ACC_PUBLIC
     * @see org.objectweb.asm.Opcodes#ACC_PROTECTED
     * @see org.objectweb.asm.Opcodes#ACC_PRIVATE
     */
    public static int makePublicAccess(int access) {
        int newAccess = access;
        newAccess = newAccess & ~Opcodes.ACC_PRIVATE;
        newAccess = newAccess & ~Opcodes.ACC_PROTECTED;
        newAccess = newAccess | Opcodes.ACC_PUBLIC;

        return newAccess;
    }

    /**
     * Changes the access flags of the class/method/field to be public and non final.
     * The method works as the {@link #makePublicAccess(int) with one difference
     * that it removes final flag from the access field.
     * 
     * @param access the access field to modify.
     * @return the access field made public and removed final flag.
     * 
     * @see #makePublicAccess(int) 
     * @see org.objectweb.asm.ClassVisitor#visit(int, int, java.lang.String, java.lang.String, java.lang.String, java.lang.String[]) 
     * @see org.objectweb.asm.ClassVisitor#visitMethod(int, java.lang.String, java.lang.String, java.lang.String, java.lang.String[])  
     * @see org.objectweb.asm.ClassVisitor#visitField(int, java.lang.String, java.lang.String, java.lang.String, java.lang.Object) 
     * @see org.objectweb.asm.Opcodes#ACC_PUBLIC
     * @see org.objectweb.asm.Opcodes#ACC_PROTECTED
     * @see org.objectweb.asm.Opcodes#ACC_PRIVATE
     * @see org.objectweb.asm.Opcodes#ACC_FINAL
     */
    public static int makePublicNonFinalAccess(int access) {
        access = makePublicAccess(access);
        access = access & ~Opcodes.ACC_FINAL;
        return access;

    }

    /**
     * Writes {@link org.objectweb.asm.Opcodes#ILOAD} instructions and like it for 
     * all of the method arguments. If the argument is object it writes
     * {@link org.objectweb.asm.Opcodes#ALOAD}, if it is a double it
     * writes {@link org.objectweb.asm.Opcodes#DLOAD} etc.
     * Also the method calculates correctly the index of argument, so if and argument is
     * long or double it assumes that they occupies two places (for more info look at
     * JVM specification).
     * 
     * @param desc the descriptor of the method whose arguments should be processed 
     *      and wrote mathing LOAD instructions.
     * @param mv the {@link org.objectweb.asm.MethodVisitor} object whose visit... should be used
     *      to write the instructions.
     * @param startIndex the starting index of method arguments. If the method is static
     *      the startIndex should be 0, otherwise should be 1.
     * 
     * @see org.objectweb.asm.Opcodes#ILOAD
     * @see org.objectweb.asm.Opcodes#ALOAD
     * @see org.objectweb.asm.Opcodes#AALOAD
     * @see org.objectweb.asm.Opcodes#FLOAD
     * @see org.objectweb.asm.Opcodes#DLOAD
     * @see org.objectweb.asm.Opcodes#LLOAD
     * @see org.objectweb.asm.Type#getOpcode(int)
     */
    public static void createLoadArgs(String desc, MethodVisitor mv, int startIndex) {
        Type[] argTypes = Type.getArgumentTypes(desc);
        int index = startIndex;
        for(int i=0;i<argTypes.length;i++) {
            int loadOpcode = argTypes[i].getOpcode(Opcodes.ILOAD);
            mv.visitIntInsn(loadOpcode, index);
            if (Type.LONG_TYPE.equals(argTypes[i]) || Type.DOUBLE_TYPE.equals(argTypes[i])) {
                index++;
            }
            index++;
        }
    }
    

}
