/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono.internal;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.Attribute;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;

/**
 * The method visitor which firstly registers the instructions of the method
 * and then could be replayed. This visitor could be used to create multiple 
 * copies of the method.
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public class RegistryMethodAdapter implements MethodVisitor {

    private static final int ANNOTATION_DEFAULT = 1;
    private static final int ANNOTATION = 2;
    private static final int PARAM_ANNOTATION = 3;
    private static final int ATTRIBUTE = 4;
    private static final int CODE = 5;
    private static final int FRAME = 6;
    private static final int INSN = 7;
    private static final int INT_INSN = 8;
    private static final int VAR_INSN = 9;
    private static final int TYPE_INSN = 10;
    private static final int FIELD_INSN = 11;
    private static final int METHOD_INSN = 12;
    private static final int JUMP_INSN = 13;
    private static final int LABEL = 14;
    private static final int LDC_INSN = 15;
    private static final int IINC_INSN = 16;
    private static final int TABLE_SWITCH_INSN = 17;
    private static final int LOOKUP_SWITCH_INSN = 18;
    private static final int MULTI_ANEW_ARRAY_INSN = 19;
    private static final int TRY_CATCH_BLOCK = 20;
    private static final int LOCAL_VARIABLE = 21;
    private static final int LINE_NUMBER = 22;
    private static final int MAXS = 23;
    private static final int END = 24;

    private List<Object[]> registry = new ArrayList<Object[]>();
    private Set<Label> labelsRegistry = new HashSet<Label>();

    /**
     * {@inheritDoc}
     */
    @Override
    public AnnotationVisitor visitAnnotationDefault() {
        RegistryAnnotationAdapter result = new RegistryAnnotationAdapter();
        registry.add(new Object[]{ANNOTATION_DEFAULT, result});
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AnnotationVisitor visitAnnotation(String desc, boolean visible) {
        RegistryAnnotationAdapter result = new RegistryAnnotationAdapter();
        registry.add(new Object[]{ANNOTATION, desc, visible, result});
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AnnotationVisitor visitParameterAnnotation(int param, String desc, boolean visible) {
        RegistryAnnotationAdapter result = new RegistryAnnotationAdapter();
        registry.add(new Object[]{PARAM_ANNOTATION, param, desc, visible, result});
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitAttribute(Attribute attr) {
        registry.add(new Object[]{ATTRIBUTE, attr});
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitCode() {
        registry.add(new Object[]{CODE});
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitFrame(int type, int nLocal, Object[] local, int nStack, Object[] stack) {
        registry.add(new Object[]{FRAME, type, nLocal, local, nStack, stack});
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitInsn(int opcode) {
        registry.add(new Object[]{INSN, opcode});
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitIntInsn(int opcode, int operand) {
        registry.add(new Object[]{INT_INSN, opcode, operand});
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitVarInsn(int opcode, int var) {
        registry.add(new Object[]{VAR_INSN, opcode, var});
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitTypeInsn(int opcode, String type) {
        registry.add(new Object[]{TYPE_INSN, opcode, type});
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitFieldInsn(int opcode, String owner, String name, String desc) {
        registry.add(new Object[]{FIELD_INSN, opcode, owner, name, desc});
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitMethodInsn(int opcode, String owner, String name, String desc) {
        registry.add(new Object[]{METHOD_INSN, opcode, owner, name, desc});
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitJumpInsn(int opcode, Label label) {
        registry.add(new Object[]{JUMP_INSN, opcode, label});
        labelsRegistry.add(label);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitLabel(Label label) {
        registry.add(new Object[]{LABEL, label});
        labelsRegistry.add(label);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitLdcInsn(Object cst) {
        registry.add(new Object[]{LDC_INSN, cst});
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitIincInsn(int var, int increment) {
        registry.add(new Object[]{IINC_INSN, var, increment});
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitTableSwitchInsn(int min, int max, Label dflt, Label[] labels) {
        registry.add(new Object[]{TABLE_SWITCH_INSN, min, max, dflt, labels});
        labelsRegistry.add(dflt);
        for(Label l : labels) {
            labelsRegistry.add(l);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitLookupSwitchInsn(Label dflt, int[] keys, Label[] labels) {
        registry.add(new Object[]{LOOKUP_SWITCH_INSN, dflt, keys, labels});
        labelsRegistry.add(dflt);
        for(Label l : labels) {
            labelsRegistry.add(l);
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitMultiANewArrayInsn(String desc, int dims) {
        registry.add(new Object[]{MULTI_ANEW_ARRAY_INSN, desc, dims});
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitTryCatchBlock(Label start, Label end, Label handler, String type) {
        registry.add(new Object[]{TRY_CATCH_BLOCK, start, end, handler, type});
        labelsRegistry.add(start);
        labelsRegistry.add(end);
        labelsRegistry.add(handler);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitLocalVariable(String name, String desc, String signature, Label start, Label end, int index) {
        registry.add(new Object[]{LOCAL_VARIABLE, name, desc, signature, start, end, index});
        labelsRegistry.add(start);
        labelsRegistry.add(end);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitLineNumber(int line, Label start) {
        registry.add(new Object[]{LINE_NUMBER, line, start});
        labelsRegistry.add(start);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitMaxs(int maxStack, int maxLocals) {
        registry.add(new Object[]{MAXS, maxStack, maxLocals});
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitEnd() {
        registry.add(new Object[]{END});
    }

    /**
     * The method used to replay the registered instructions. This method could be 
     * used for multiple times.
     * 
     * @param visitor the next visitor in the chain (where the instructions should be written).
     */
    public void replay(MethodVisitor visitor) {
        Map<Label, Label> labelsMap = new Hashtable<Label, Label>();
        for(Label l : labelsRegistry) {
            labelsMap.put(l, new Label());
        }


        for(Object[] obj : registry) {
            switch((Integer)obj[0]) {
                case ANNOTATION_DEFAULT:
                    AnnotationVisitor av_ann_def = visitor.visitAnnotationDefault();
                    if (av_ann_def != null) {
                        ((RegistryAnnotationAdapter)obj[1]).replay(av_ann_def);
                    }
                    break;
                case ANNOTATION:
                    AnnotationVisitor av_ann = visitor.visitAnnotation((String)obj[1], (Boolean)obj[2]);
                    if (av_ann != null) {
                        ((RegistryAnnotationAdapter)obj[3]).replay(av_ann);
                    }
                    break;
                case PARAM_ANNOTATION:
                    AnnotationVisitor av_param_ann = visitor.visitParameterAnnotation((Integer)obj[1], (String)obj[2], (Boolean)obj[3]);
                    if (av_param_ann != null) {
                        ((RegistryAnnotationAdapter)obj[4]).replay(av_param_ann);
                    }
                    break;
                case ATTRIBUTE: visitor.visitAttribute((Attribute)obj[1]); break;
                case CODE: visitor.visitCode(); break;
                case FRAME: visitor.visitFrame(
                        (Integer) obj[1],
                        (Integer) obj[2],
                        (Object[]) obj[3],
                        (Integer) obj[4],
                        (Object[]) obj[5]); break;
                case INSN: visitor.visitInsn((Integer) obj[1]); break;
                case INT_INSN: visitor.visitIntInsn((Integer) obj[1], (Integer) obj[2]); break;
                case VAR_INSN: visitor.visitVarInsn((Integer) obj[1], (Integer) obj[2]); break;
                case TYPE_INSN: visitor.visitTypeInsn((Integer) obj[1], (String) obj[2]); break;
                case FIELD_INSN: visitor.visitFieldInsn(
                        (Integer) obj[1],
                        (String) obj[2],
                        (String) obj[3],
                        (String) obj[4]); break;
                case METHOD_INSN: visitor.visitMethodInsn(
                        (Integer) obj[1],
                        (String) obj[2],
                        (String) obj[3],
                        (String) obj[4]); break;
                case JUMP_INSN: visitor.visitJumpInsn(
                        (Integer) obj[1],
                        getLabel(labelsMap, obj[2])); break;
                case LABEL: visitor.visitLabel(getLabel(labelsMap, obj[1])); break;
                case LDC_INSN: visitor.visitLdcInsn(obj[1]); break;
                case IINC_INSN: visitor.visitIincInsn((Integer)obj[1], (Integer)obj[2]); break;
                case TABLE_SWITCH_INSN:
                    visitor.visitTableSwitchInsn(
                        (Integer)obj[1],
                        (Integer)obj[2],
                        getLabel(labelsMap, obj[3]),
                        getLabels(labelsMap, obj[4])); break;
                case LOOKUP_SWITCH_INSN: visitor.visitLookupSwitchInsn(
                        getLabel(labelsMap, obj[1]),
                        (int[])obj[2],
                        getLabels(labelsMap, obj[3])); break;
                case MULTI_ANEW_ARRAY_INSN: visitor.visitMultiANewArrayInsn(
                        (String)obj[1],
                        (Integer)obj[2]); break;
                case TRY_CATCH_BLOCK: visitor.visitTryCatchBlock(
                        getLabel(labelsMap, obj[1]),
                        getLabel(labelsMap, obj[2]),
                        getLabel(labelsMap, obj[3]),
                        (String)obj[4]); break;
                case LOCAL_VARIABLE: visitor.visitLocalVariable(
                        (String)obj[1],
                        (String)obj[2],
                        (String)obj[3],
                        getLabel(labelsMap, obj[4]),
                        getLabel(labelsMap, obj[5]),
                        (Integer)obj[6]); break;
                case LINE_NUMBER: visitor.visitLineNumber(
                        (Integer)obj[1],
                        getLabel(labelsMap, obj[2])); break;
                case MAXS: visitor.visitMaxs(
                        (Integer)obj[1],
                        (Integer)obj[2]); break;
                case END: visitor.visitEnd(); break;
            }
        }
    }

    private Label getLabel(Map<Label, Label> map, Object obj) {
        return map.get((Label) obj);
    }

    private Label[] getLabels(Map<Label, Label> map, Object obj) {
        Label[] labels = (Label[])obj;
        Label[] result = new Label[labels.length];

        for(int i=0;i<labels.length;i++) {
            result[i] = map.get(labels[i]);
        }

        return result;
    }

}
