/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono.internal;

import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.Attribute;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodAdapter;
import org.objectweb.asm.MethodVisitor;

/**
 * The method visitor which have a state. This visitor could be used to
 * transform the method depending of the state machine. The state machine
 * does not persist the state if in the descendant it is not provided.
 * All of the methods resets the state machine. To avoid this override the method
 * and ignore resetting the state.
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public class StatefullMethodAdapter extends MethodAdapter {
    
    private boolean isResetting = false;
    
    /**
     * The initial state of the state machine.
     */
    protected static final int STATE_NONE = 0;
    
    /*
     * The current state if the state machine.
     */
    protected int state = STATE_NONE;

    /**
     * The constructor used to set the next method visitor in the chain.
     * 
     * @param mv the next method visitor in the chain.
     */
    public StatefullMethodAdapter(MethodVisitor mv) {
        super(mv);
    }
    
    /**
     * The method used by descendants to reset the state machine.
     */
    protected void performReset() {
        
    }
    
    /**
     * Resets the state machine and performs the custom actions.
     */
    private void resetState() {
        if (isResetting) {
            return;
        }
        isResetting = true;
        performReset();
        state = STATE_NONE;
        isResetting = false;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AnnotationVisitor visitAnnotation(String string, boolean bln) {
        resetState();
        return super.visitAnnotation(string, bln);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AnnotationVisitor visitAnnotationDefault() {
        resetState();
        return super.visitAnnotationDefault();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitAttribute(Attribute atrbt) {
        resetState();
        super.visitAttribute(atrbt);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitCode() {
        resetState();
        super.visitCode();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitEnd() {
        resetState();
        super.visitEnd();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitFieldInsn(int i, String string, String string1, String string2) {
        resetState();
        super.visitFieldInsn(i, string, string1, string2);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitFrame(int i, int i1, Object[] os, int i2, Object[] os1) {
        resetState();
        super.visitFrame(i, i1, os, i2, os1);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitIincInsn(int i, int i1) {
        resetState();
        super.visitIincInsn(i, i1);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitInsn(int i) {
        resetState();
        super.visitInsn(i);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitIntInsn(int i, int i1) {
        resetState();
        super.visitIntInsn(i, i1);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitJumpInsn(int i, Label label) {
        resetState();
        super.visitJumpInsn(i, label);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitLabel(Label label) {
        resetState();
        super.visitLabel(label);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitLdcInsn(Object o) {
        resetState();
        super.visitLdcInsn(o);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitLineNumber(int i, Label label) {
        resetState();
        super.visitLineNumber(i, label);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitLocalVariable(String string, String string1, String string2, Label label, Label label1, int i) {
        resetState();
        super.visitLocalVariable(string, string1, string2, label, label1, i);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitLookupSwitchInsn(Label label, int[] ints, Label[] labels) {
        resetState();
        super.visitLookupSwitchInsn(label, ints, labels);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitMaxs(int i, int i1) {
        resetState();
        super.visitMaxs(i, i1);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitMethodInsn(int i, String string, String string1, String string2) {
        resetState();
        super.visitMethodInsn(i, string, string1, string2);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitMultiANewArrayInsn(String string, int i) {
        resetState();
        super.visitMultiANewArrayInsn(string, i);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public AnnotationVisitor visitParameterAnnotation(int i, String string, boolean bln) {
        resetState();
        return super.visitParameterAnnotation(i, string, bln);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitTableSwitchInsn(int i, int i1, Label label, Label[] labels) {
        resetState();
        super.visitTableSwitchInsn(i, i1, label, labels);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitTryCatchBlock(Label label, Label label1, Label label2, String string) {
        resetState();
        super.visitTryCatchBlock(label, label1, label2, string);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitTypeInsn(int i, String string) {
        resetState();
        super.visitTypeInsn(i, string);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void visitVarInsn(int i, int i1) {
        resetState();
        super.visitVarInsn(i, i1);
    }

    
    
}
