/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono.internal.modifiers;

import org.objectweb.asm.Type;
import java.util.Calendar;
import com.qitsoft.qimono.internal.ClassModifierAdapter;
import com.qitsoft.qimono.internal.QimonoClassLoader;
import com.qitsoft.qimono.internal.substitutors.DateSubstitutor;
import java.util.GregorianCalendar;
import org.objectweb.asm.MethodAdapter;
import org.objectweb.asm.MethodVisitor;

import static org.objectweb.asm.Opcodes.*;

/**
 * The Class Modifier used to intercept the {@link java.util.Calendar}. It intercepts instantiations
 * and usages of {@link java.util.Calendar#getInstance()}, {@link java.util.Calendar#getInstance(java.util.Locale) },
 * {@link java.util.Calendar#getInstance(java.util.TimeZone) } and {@link java.util.Calendar#getInstance(java.util.TimeZone, java.util.Locale) }
 * to set the pointing date to installed by user through {@link com.qitsoft.qimono.internal.substitutors.DateSubstitutor}.
 * It also intercepts the usages of {@link java.util.GregorianCalendar}.
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public class CalendarInterceptModifier extends ClassModifierAdapter {
    
    /**
     * Configures {@link com.qitsoft.qimono.internal.QimonoClassLoader} to use
     * this class modifier.
     * 
     * @param classloader the class loader to configure.
     */
    public static void configureClassloader(QimonoClassLoader classloader) {
        classloader.addClassModifier(new CalendarInterceptModifier());
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public MethodVisitor visitMethod(int access, final String name, String descriptor, String signature, String[] exceptions) {
        return new MethodAdapter(super.visitMethod(access, name, descriptor, signature, exceptions)) {

            private static final int STATE_NONE = 0;
            private static final int STATE_NEW = 1;

            private int state = STATE_NONE;

            private boolean isInitMethod() {
                return "<init>".equals(name);
            }
            
            @Override
            public void visitTypeInsn(int opcode, String type) {
                if (opcode == NEW
                        && (Type.getInternalName(Calendar.class).equals(type)
                        || Type.getInternalName(GregorianCalendar.class).equals(type))) {
                    
                    state = STATE_NEW;
                }
                super.visitTypeInsn(opcode, type);
            }
            
            @Override
            public void visitMethodInsn(int opcode, String owner, String name, String descriptor) {
                if (opcode == INVOKESTATIC
                        && (Type.getInternalName(Calendar.class).equals(owner)
                        || Type.getInternalName(GregorianCalendar.class).equals(owner))
                        && "getInstance".equals(name)) {
                    
                    super.visitMethodInsn(INVOKESTATIC, Type.getInternalName(DateSubstitutor.class), "getCalendarInstance", descriptor);
                } else if ((state == STATE_NEW || isInitMethod())
                        && opcode == INVOKESPECIAL
                        && (Type.getInternalName(Calendar.class).equals(owner) || Type.getInternalName(GregorianCalendar.class).equals(owner))
                        && "<init>".equals(name)
                        && ("()V".equals(descriptor)
                        || "(Ljava/util/Locale;)V".equals(descriptor))) {
                    
                    super.visitMethodInsn(opcode, owner, name, descriptor);
                    if (isInitMethod()) {
                        super.visitVarInsn(ALOAD, 0);
                    }
                    super.visitMethodInsn(INVOKESTATIC, Type.getInternalName(DateSubstitutor.class), "calendarInit", "("+Type.getDescriptor(Calendar.class)+")"+Type.getDescriptor(Calendar.class));
                    if (Type.getInternalName(GregorianCalendar.class).equals(owner)) {
                        super.visitTypeInsn(CHECKCAST, Type.getInternalName(GregorianCalendar.class));
                    }
                    state = STATE_NONE;
                } else if ((state == STATE_NEW || isInitMethod())
                        && opcode == INVOKESPECIAL
                        && (Type.getInternalName(Calendar.class).equals(owner) || Type.getInternalName(GregorianCalendar.class).equals(owner))
                        && "<init>".equals(name)
                        && ("(Ljava/util/TimeZone;)V".equals(descriptor)
                        || "(Ljava/util/TimeZone;Ljava/util/Locale;)V".equals(descriptor))) {
                    
                    super.visitMethodInsn(opcode, owner, name, descriptor);
                    if (isInitMethod()) {
                        super.visitVarInsn(ALOAD, 0);
                    }
                    super.visitMethodInsn(INVOKESTATIC, Type.getInternalName(DateSubstitutor.class), "calendarInitWithTimeZone", "("+Type.getDescriptor(Calendar.class)+")"+Type.getDescriptor(Calendar.class));
                    if (Type.getInternalName(GregorianCalendar.class).equals(owner)) {
                        super.visitTypeInsn(CHECKCAST, Type.getInternalName(GregorianCalendar.class));
                    }
                    state = STATE_NONE;
                } else {
                    super.visitMethodInsn(opcode, owner, name, descriptor);
                }
            }
            
        };
    }
    
    
}
