/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono.internal.modifiers;

import com.qitsoft.qimono.internal.ClassModifierAdapter;
import com.qitsoft.qimono.internal.QimonoClassLoader;
import com.qitsoft.qimono.internal.StatefullMethodAdapter;
import com.qitsoft.qimono.internal.substitutors.DateSubstitutor;
import java.util.Date;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;

/**
 * Modifies the class to intercept {@link java.util.Date} usage. It intercept the
 * instantiations of {@link java.util.Date} and extending of date to set the date
 * installed by user through {@link com.qitsoft.qimono.internal.substitutors.DateSubstitutor}.
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public class DateInterceptModifier extends ClassModifierAdapter {

    private static DateInterceptModifier modifier;
    
    /**
     * Configures the {@link com.qitsoft.qimono.internal.QimonoClassLoader}. It
     * adds itself as a class modifier.
     * 
     * @param classloader the class loader to configure.
     */
    public static void configureClassloader(QimonoClassLoader classloader) {
        modifier = new DateInterceptModifier();
        classloader.addClassModifier(modifier);
    }

    /**
     * {@inheritDoc 
     */
    @Override
    public MethodVisitor visitMethod(int access, final String name, String description, String signature, String[] exceptions) {
        return new StatefullMethodAdapter(super.visitMethod(access, name, description, signature, exceptions)) {

            private static final int STATE_NEW = 1;
            private static final int STATE_DUP = 2;
           
            private boolean isInitMethod() {
                return "<init>".equals(name);
            }
            
            @Override
            public void visitTypeInsn(int opcode, String type) {
                if (state == STATE_NONE && Type.getInternalName(Date.class).equals(type)) {
                    state = STATE_NEW;
                } else {
                    super.visitTypeInsn(opcode, type);
                }
            }

            @Override
            public void visitInsn(int opcode) {
                if (state == STATE_NEW && opcode == Opcodes.DUP) {
                    state = STATE_DUP;
                } else {
                    super.visitInsn(opcode);
                }
            }

            @Override
            public void visitMethodInsn(int opcode, String owner, String method, String desc) {
                if (state == STATE_DUP
                        && opcode == Opcodes.INVOKESPECIAL 
                        && Type.getInternalName(Date.class).equals(owner)
                        && "<init>".equals(method)
                        && "()V".equals(desc)) {
                    
                    state = STATE_NONE;
                    super.visitMethodInsn(Opcodes.INVOKESTATIC, Type.getInternalName(DateSubstitutor.class), "newDate", "()"+Type.getDescriptor(Date.class));
                } else if (isInitMethod()
                        && state == STATE_NONE
                        && opcode == Opcodes.INVOKESPECIAL 
                        && Type.getInternalName(Date.class).equals(owner)
                        && "<init>".equals(method)
                        && "()V".equals(desc)) {
                    
                    super.visitMethodInsn(opcode, owner, method, desc);
                    super.visitVarInsn(Opcodes.ALOAD, 0);
                    super.visitMethodInsn(Opcodes.INVOKESTATIC, Type.getInternalName(DateSubstitutor.class), "initDate", "("+Type.getDescriptor(Date.class)+")"+Type.getDescriptor(Date.class));
                } else {
                    super.visitMethodInsn(opcode, owner, method, desc);
                }
            }

            @Override
            protected void performReset() {
                switch(state) {
                    case STATE_NEW: 
                        super.visitTypeInsn(Opcodes.NEW, Type.getInternalName(Date.class));
                        break;
                    case STATE_DUP:
                        super.visitTypeInsn(Opcodes.NEW, Type.getInternalName(Date.class));
                        super.visitInsn(Opcodes.DUP);
                        break;
                }
                super.performReset();
            }
            
        };
    }
    
    
    
}
