/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono.internal.modifiers;

import com.qitsoft.qimono.internal.ClassModifierUtils;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;

/**
 * Modifies the class removing final methods and making the class non final.
 * This modifier is required for {@link org.mockito.Mockito} to be able to
 * mock the final classes and final methods.
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public class FinalMethodsModifier extends MethodsCoppierModifier {

    /**
     * {@inheritDoc }
     */
    @Override
    public void visit(int version, int access, String name, String signature, String superName, String[] interfaces) {
        access = ClassModifierUtils.makePublicNonFinalAccess(access);
        super.visit(version, access, name, signature, superName, interfaces);
    }

    /**
     * {@inheritDoc }
     */
    @Override
    protected boolean isMethodToCopy(int modifiers, String name, String desc, String signature, String[] exceptions) {
        return (modifiers & Opcodes.ACC_FINAL) != 0
                && (modifiers & Opcodes.ACC_STATIC) == 0
                && !name.endsWith("$QIMONO$ORIGINAL")
                && !name.equals("<init>");
    }

    /**
     * {@inheritDoc }
     */
    @Override
    protected MethodVisitor getOriginalMethodVisitor(MethodVisitor visitor, ClassDecl classDecl, MethodDecl method) {
        return new FinalMethodVisitor(visitor, classDecl, method);
    }

    /**
     * {@inheritDoc }
     */
    @Override
    protected int modifyCopyMethodAccess(int modifiers) {
        return ClassModifierUtils.makePublicNonFinalAccess(modifiers);
    }

    /**
     * {@inheritDoc }
     */
    @Override
    protected int modifyOriginalMethodAccess(int modifiers) {
        return ClassModifierUtils.makePublicAccess(modifiers);
    }

    /**
     * The Method visitor used to modify the original method. It removes the content of it
     * and replaces with the invocation of the method with original content but non final.
     */
    private class FinalMethodVisitor extends OriginalMethodVisitor {

        public FinalMethodVisitor(MethodVisitor mv, ClassDecl classDecl, MethodDecl methodDesc) {
            super(mv, classDecl, methodDesc);
        }

        @Override
        public void visitCode() {
            mv.visitCode();

            mv.visitIntInsn(Opcodes.ALOAD, 0);

            Type[] argTypes = Type.getArgumentTypes(methodDecl.desc);
            int index = 1;
            for(int i=0;i<argTypes.length;i++) {
                int loadOpcode = argTypes[i].getOpcode(Opcodes.ILOAD);
                mv.visitIntInsn(loadOpcode, index);
                if (Type.LONG_TYPE.equals(argTypes[i]) || Type.DOUBLE_TYPE.equals(argTypes[i])) {
                    index++;
                }
                index++;
            }

            mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, classDecl.name, modifyCopyMethodName(methodDecl.name), methodDecl.desc);

            Type returnType = Type.getReturnType(methodDecl.desc);
            mv.visitInsn(returnType.getOpcode(Opcodes.IRETURN));
            mv.visitMaxs(0, 0);
        }
    }

}
