/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono.internal.modifiers;

import com.qitsoft.qimono.internal.ClassModifierAdapter;
import com.qitsoft.qimono.internal.ClassModifierUtils;
import com.qitsoft.qimono.internal.RegistryMethodAdapter;
import java.util.ArrayList;
import java.util.List;
import org.objectweb.asm.AnnotationVisitor;
import org.objectweb.asm.Attribute;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodAdapter;
import org.objectweb.asm.MethodVisitor;

/**
 * The modifier used to copy the methods. It is used to copy the content of the
 * method into a new one with the special name and to replace the body of the
 * original method with  invocation to the new methods created.
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public abstract class MethodsCoppierModifier extends ClassModifierAdapter {

    /**
     * The methods market to copy.
     */
    protected List<MethodDecl> methods = new ArrayList<MethodDecl>();
    
    /**
     * The current class declaration.
     */
    protected ClassDecl classDecl;

    /**
     * {@inheritDoc }
     */
    @Override
    public void visit(int version, int access, String name, String signature, String superName, String[] interfaces) {
        super.visit(version, access, name, signature, superName, interfaces);
        methods.clear();
        classDecl = new ClassDecl(access, name, signature, superName, interfaces);
    }

    /**
     * Returns if the method should be copied. This is the abstract method and should
     * be used by descendants to specify if the method should be copied.
     * 
     * @param modifiers the access modifiers of the method.
     * @param name the name of the method.
     * @param desc the descriptor of the method.
     * @param signature the method's signature.
     * @param exceptions the exceptions thrown by the method.
     * @return true if the method should be copied.
     */
    protected abstract boolean isMethodToCopy(int modifiers, String name, String desc, String signature, String[] exceptions);
    
    /**
     * Modifies the access modifiers of the original method. By default
     * it makes the method to be public only (replaces private, protected and package visibilities).
     * @param modifiers the initial access modifier.
     * @return the modified access modifiers.
     */
    protected int modifyOriginalMethodAccess(int modifiers) {
        return ClassModifierUtils.makePublicAccess(modifiers);
    }
    
    /**
     * Returns the access modifiers of the copied method. By default it
     * makes the method to be public (replacing private, protected or package visibilities).
     * 
     * @param modifiers the access modifiers of the original method.
     * @return the access modifiers to be applied to copied method.
     */
    protected int modifyCopyMethodAccess(int modifiers) {
        return ClassModifierUtils.makePublicAccess(modifiers);
    }
    
    /**
     * Returns the copied method name. By default it appends to the original method name
     * "$QIMONO$ORIGINAL".
     * 
     * @param name the original method name.
     * @return the name to be applied to copied method.
     */
    protected String modifyCopyMethodName(String name) {
        return name + "$QIMONO$ORIGINAL";
    }
    
    /**
     * Returns the method visitor for the original method. By default it removes the body.
     * It uses the {@link OriginalMethodVisitor} modifier.
     * 
     * @param visitor the next method visitor in the chain.
     * @param classDecl the {@link ClassDecl} of the current class.
     * @param method the {@link MethodDecl} of the original method to modify.
     * @return the method visitor used to modify the content of the original method.
     */
    protected MethodVisitor getOriginalMethodVisitor(MethodVisitor visitor, ClassDecl classDecl, MethodDecl method) {
        return new OriginalMethodVisitor(visitor, classDecl, method);
    }
    
    /**
     * The Method visitor used to modify the body of the copied method. by default it 
     * only removes the annotations and the body is the same as of the original.
     * @param visitor the next method visitor in the chain.
     * @param classDecl the current class {@link ClassDecl}.
     * @param method the current method's {@link MethodDecl}.
     * @return the class visitor used to modify the body of the copied method.
     */
    protected MethodVisitor getCopyMethodVisitor(MethodVisitor visitor, ClassDecl classDecl, MethodDecl method) {
        return new CopyMethodVisitor(visitor, classDecl, method);
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public MethodVisitor visitMethod(int modifiers, String name, String desc, String signature, String[] exceptions) {
        if (isMethodToCopy(modifiers, name, desc, signature, exceptions)) {
            RegistryMethodAdapter adapter = new RegistryMethodAdapter();
            methods.add(new MethodDecl(modifiers, name, desc, signature, exceptions, adapter));
            return adapter;
        } else {
            return super.visitMethod(modifiers, name, desc, signature, exceptions);
        }
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public void visitEnd() {
        for(MethodDecl method : methods) {
            // add original method
            MethodVisitor originalVisitor = super.visitMethod(
                    modifyOriginalMethodAccess(method.modifiers),
                    method.name, method.desc, method.signature, method.exceptions);

            method.adapter.replay(getOriginalMethodVisitor(originalVisitor, classDecl, method));

            // add final method
            MethodVisitor copyVisitor = super.visitMethod(
                    modifyCopyMethodAccess(method.modifiers),
                    modifyCopyMethodName(method.name), method.desc, method.signature, method.exceptions);

            method.adapter.replay(getCopyMethodVisitor(copyVisitor, classDecl, method));
        }

        super.visitEnd();
    }

    /**
     * The class used to modify the body of the original method. All of
     * the method do nothing (removes the original content).
     */
    protected static class OriginalMethodVisitor extends MethodAdapter {

        /**
         * The current class.
         */
        public final ClassDecl classDecl;
        
        /**
         * The current method.
         */
        public final MethodDecl methodDecl;

        /**
         * Base constructor used to initialize the method visitor.
         * @param mv the next method visitor in the chain.
         * @param classDecl the current class.
         * @param methodDesc the current method.
         */
        public OriginalMethodVisitor(MethodVisitor mv, ClassDecl classDecl, MethodDecl methodDesc) {
            super(mv);
            this.classDecl = classDecl;
            this.methodDecl = methodDesc;
        }

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitFieldInsn(int i, String string, String string1, String string2) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitFrame(int i, int i1, Object[] os, int i2, Object[] os1) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitIincInsn(int i, int i1) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitInsn(int i) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitIntInsn(int i, int i1) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitJumpInsn(int i, Label label) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitLabel(Label label) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitLdcInsn(Object o) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitLineNumber(int i, Label label) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitLocalVariable(String string, String string1, String string2, Label label, Label label1, int i) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitLookupSwitchInsn(Label label, int[] ints, Label[] labels) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitMethodInsn(int i, String string, String string1, String string2) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitMultiANewArrayInsn(String string, int i) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitTableSwitchInsn(int i, int i1, Label label, Label[] labels) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitTryCatchBlock(Label label, Label label1, Label label2, String string) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitTypeInsn(int i, String string) {}

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitVarInsn(int i, int i1) {}

    }

    /**
     * Method visitor used to modify the copied method. It removes only the
     * annotations. The rest remains untouched.
     */
    protected static class CopyMethodVisitor extends MethodAdapter {

        /**
         * Current class.
         */
        public final ClassDecl classDecl;
        
        /**
         * Current method.
         */
        public final MethodDecl methodDesc;

        /**
         * Base constructor used to instantiate and initialize this method visitor.
         * 
         * @param mv the next method visitor in the chain.
         * @param classDecl current class.
         * @param methodDesc current method.
         */
        public CopyMethodVisitor(MethodVisitor mv, ClassDecl classDecl, MethodDecl methodDesc) {
            super(mv);
            this.classDecl = classDecl;
            this.methodDesc = methodDesc;
        }

        /**
         * {@inheritDoc }
         */
        @Override
        public AnnotationVisitor visitAnnotation(String string, boolean bln) {
            return null;
        }

        /**
         * {@inheritDoc }
         */
        @Override
        public AnnotationVisitor visitAnnotationDefault() {
            return null;
        }

        /**
         * {@inheritDoc }
         */
        @Override
        public void visitAttribute(Attribute atrbt) {
            return;
        }

    }


}
