/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono.internal.modifiers;

import com.qitsoft.qimono.internal.ClassModifierAdapter;
import com.qitsoft.qimono.internal.ClassModifierUtils;
import com.qitsoft.qimono.internal.QimonoClassLoader;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;

/**
 * Class Modifier used to make all static method to be mockable.
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public class StaticMethodsModifier extends MethodsCoppierModifier {

    /**
     * Configures the class loader to use this modifier. Also it adds the
     * synthetic class modifier.
     * 
     * @param classloader the class loader to modify.
     */
    public static void configureClassloader(QimonoClassLoader classloader) {
        StaticMethodsModifier modifier = new StaticMethodsModifier();
        classloader.addClassModifier(modifier);
        classloader.addSyntheticClassModifier(new StaticMethodsAdderVisitor(modifier));
    }

    /**
     * {@inheritDoc }
     */
    @Override
    protected boolean isMethodToCopy(int modifiers, String name, String desc, String signature, String[] exceptions) {
        return (modifiers & Opcodes.ACC_STATIC) != 0
                && !name.endsWith("$QIMONO$ORIGINAL")
                && !name.equals("<clinit>");
    }

    /**
     * {@inheritDoc }
     */
    @Override
    protected MethodVisitor getOriginalMethodVisitor(MethodVisitor visitor, ClassDecl classDecl, MethodDecl method) {
        return new StaticMethodVisitor(visitor, classDecl, method);
    }
    
    private class StaticMethodVisitor extends OriginalMethodVisitor {

        public StaticMethodVisitor(MethodVisitor mv, ClassDecl classDecl, MethodDecl methodDesc) {
            super(mv, classDecl, methodDesc);
        }

        @Override
        public void visitCode() {
            mv.visitCode();

            mv.visitLdcInsn(Type.getType("L"+classDecl.name+";"));
            mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, Type.getInternalName(Class.class), "getClassLoader", "()Ljava/lang/ClassLoader;");
            mv.visitTypeInsn(Opcodes.CHECKCAST, Type.getInternalName(QimonoClassLoader.class));
            mv.visitLdcInsn(Type.getType("L"+classDecl.name+";"));
            mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, Type.getInternalName(QimonoClassLoader.class), "getSyntheticMock", "(Ljava/lang/Class;)Ljava/lang/Object;");
            mv.visitTypeInsn(Opcodes.CHECKCAST, "qimono/synthetic/"+classDecl.name);
            ClassModifierUtils.createLoadArgs(methodDecl.desc, mv, 0);

            mv.visitMethodInsn(Opcodes.INVOKEVIRTUAL, "qimono/synthetic/"+classDecl.name, methodDecl.name, methodDecl.desc);

            Type returnType = Type.getReturnType(methodDecl.desc);
            mv.visitInsn(returnType.getOpcode(Opcodes.IRETURN));
            mv.visitMaxs(0, 0);
        }


    }

    private static class StaticMethodsAdderVisitor extends ClassModifierAdapter {

        private StaticMethodsModifier modifier;

        public StaticMethodsAdderVisitor(StaticMethodsModifier modifier) {
            this.modifier = modifier;
        }
        
        @Override
        public void visitEnd() {
            for(MethodDecl method : modifier.methods) {
                MethodVisitor mv = visitMethod(Opcodes.ACC_PUBLIC, method.name, method.desc, method.signature, method.exceptions);
                createSyntheticMethod(mv, method);
            }
            
            super.visitEnd();
        }

        private void createSyntheticMethod(MethodVisitor mv, MethodDecl method) {
            mv.visitCode();
            
            ClassModifierUtils.createLoadArgs(method.desc, mv, 1);

            mv.visitMethodInsn(Opcodes.INVOKESTATIC, modifier.classDecl.name, modifier.modifyCopyMethodName(method.name), method.desc);

            Type returnType = Type.getReturnType(method.desc);
            mv.visitInsn(returnType.getOpcode(Opcodes.IRETURN));
            mv.visitMaxs(0, 0);
            mv.visitEnd();
        }
    }


}
