/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono.internal.substitutors;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.Locale;
import java.util.TimeZone;

/**
 * The class used to substitute the date and time. This class holds the date and time
 * and when it is used to retrieve the current time it returns the set on with time shift
 * since it was installed. This class is thread safe.
 * 
 * You can set specified fields, the rest are used of the current environment. 
 * If no fields were set it uses the current time.
 * 
 * You can setup the date and time in several steps, eg:
 * 
 * <pre><code> 
 * DateSubstitutor.setDate(2010, 01, 12); // Sets 12th of January 2010, time points to the current env time.
 * DateSubstitutor.setTime(15, 23, 34); // Sets the time 15:23:34
 * DateSubstitutor.setTimeZone(TimeZone.getTimeZone("GMT+2")); // Sets the time zone of GMT+0200
 * // Currently the date and time is 12th of January 2010 15:23:34 GMT+0200 
 * // The milliseconds remains of the current environment
 * </code></pre>
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public class DateSubstitutor {
    
    private static ThreadLocal<Integer> year = new ThreadLocal<Integer>();
    private static ThreadLocal<Integer> month = new ThreadLocal<Integer>();
    private static ThreadLocal<Integer> date = new ThreadLocal<Integer>();
    private static ThreadLocal<Integer> hour = new ThreadLocal<Integer>();
    private static ThreadLocal<Integer> minute = new ThreadLocal<Integer>();
    private static ThreadLocal<Integer> second = new ThreadLocal<Integer>();
    private static ThreadLocal<Integer> millisecond = new ThreadLocal<Integer>();
    private static ThreadLocal<TimeZone> timezone = new ThreadLocal<TimeZone>();
    private static ThreadLocal<Long> setupTime = new ThreadLocal<Long>();

    /**
     * Constructs the date. 
     * @return the {@link java.util.Date} with the date provided by the user.
     */
    public static Date newDate() {
        return getCalendarInstance().getTime();
    }
    
    /**
     * Returns the time in milliseconds since 1970-01-01 00:00:00 of the provided 
     * time.
     * 
     * @return the time provided by the user.
     */
    public static long currentTimeMillis() {
        return getCalendarInstance().getTimeInMillis();
    }

    /**
     * Returns the {@link java.util.Calendar} instance with provided date and time.
     * 
     * @return the Calendar instance with provided time.
     */
    public static Calendar getCalendarInstance() {
        return getCalendarInstance(null, null);
    }

    /**
     * Returns the {@link java.util.Calendar} instance with provided date and time.
     * This method transforms the provided time to the one in the specified time zone.
     * 
     * @param aTimeZone the time zone of the new calendar instance.
     * @return the calendar with provided date and time.
     */
    public static Calendar getCalendarInstance(TimeZone aTimeZone) {
        return getCalendarInstance(aTimeZone, null);
    }

    /**
     * Returns the {@link java.util.Calendar} instance with provided date and time.
     * 
     * @param locale the locale of the new calendar instance.
     * @return the new calendar instance.
     */
    public static Calendar getCalendarInstance(Locale locale) {
        return getCalendarInstance(null, locale);
    }
    
    /**
     * Returns the {@link java.util.Calendar} instance with provided date and time.
     * This method transforms the provided time to the one in the specified time zone.
     * 
     * @param aTimeZone the time zone of the new calendar instance.
     * @param locale the locale of the new calendar instance.
     * @return the calendar with provided date and time.
     */
    public static Calendar getCalendarInstance(TimeZone aTimeZone, Locale locale) {
        Calendar calendar;
        if (locale == null) {
            locale = Locale.getDefault();
        }
        if (timezone.get() == null) {
            calendar = new GregorianCalendar();
        } else {
            calendar = new GregorianCalendar(timezone.get());
        }
        
        if (year.get() != null) {
            calendar.set(Calendar.YEAR, year.get());
        }
        if (month.get() != null) {
            calendar.set(Calendar.MONTH, month.get());
        }
        if (date.get() != null) {
            calendar.set(Calendar.DATE, date.get());
        }
        if (hour.get() != null) {
            calendar.set(Calendar.HOUR_OF_DAY, hour.get());
        }
        if (minute.get() != null) {
            calendar.set(Calendar.MINUTE, minute.get());
        }
        if (second.get() != null) {
            calendar.set(Calendar.SECOND, second.get());
        }
        if (millisecond.get() != null) {
            calendar.set(Calendar.MILLISECOND, millisecond.get());
        }
        if (setupTime.get() != null) {
            long currentTime = System.currentTimeMillis();
            int ellapsedTime = (int) (currentTime - setupTime.get());
            calendar.add(Calendar.MILLISECOND, ellapsedTime);
        }
        
        if (aTimeZone != null) {
            calendar = shiftByTimeZone(new GregorianCalendar(aTimeZone, locale), calendar);
        }
        
        return calendar;
    }

    public static <T extends Calendar> T shiftByTimeZone(T calendar, Calendar originalCalendar) {
        calendar.setTimeInMillis(originalCalendar.getTimeInMillis());
        return calendar;
    }

    /**
     * Performs initialization of the {@link java.util.Calendar} accordingly
     * with the date and time provided.
     * 
     * @param <T> the subtype of calendar.
     * @param calendar the calendar to initialize.
     * @return the instance of calendar initialized.
     */
    public static <T extends Calendar> T calendarInit(T calendar) {
        return shiftByTimeZone(calendar, getCalendarInstance());
    }
    
    /**
     * Initialize the calendar instance which was constructed with specifying time zone.
     * 
     * @param <T> the subtype of the calendar.
     * @param calendar the calendar to initialize.
     * @return the initialized calendar.
     */
    public static <T extends Calendar> T calendarInitWithTimeZone(T calendar) {
        return shiftByTimeZone(calendar, getCalendarInstance());
    }
    
    /**
     * Resets all the fields to the current system time.
     */
    public static void reset() {
        year.remove();
        month.remove();
        date.remove();
        hour.remove();
        minute.remove();
        second.remove();
        millisecond.remove();
        timezone.remove();
        setupTime.remove();
    }

    /**
     * Set the date and time. The time zone is ignored.
     * 
     * @param aYear the year to set.
     * @param aMonth the month to set. January is 1, February is 2 etc.
     * @param aDate the date of month.
     * @param aHour the hour of day in 24 hours format.
     * @param aMinute the minutes.
     * @param aSecond the seconds.
     * @param aMillisecond the milliseconds.
     */
    public static void setDateTime(int aYear, int aMonth, int aDate, int aHour, int aMinute, int aSecond, int aMillisecond) {
        year.set(aYear);
        month.set(aMonth - 1);
        date.set(aDate);
        hour.set(aHour);
        minute.set(aMinute);
        second.set(aSecond);
        millisecond.set(aMillisecond);
        setupCurrentTime();
        
    }

    /**
     * Set the date and time. The milliseconds and time zone is ignored.
     * 
     * @param aYear the year.
     * @param aMonth the month of the year. January is 1, February is 2 etc.
     * @param aDate the day of month.
     * @param aHour the hour of the day in 24 format.
     * @param aMinute the minutes.
     * @param aSecond the seconds.
     */
    public static void setDateTime(int aYear, int aMonth, int aDate, int aHour, int aMinute, int aSecond) {
        year.set(aYear);
        month.set(aMonth - 1);
        date.set(aDate);
        hour.set(aHour);
        minute.set(aMinute);
        second.set(aSecond);
        setupCurrentTime();
    }

    /**
     * Sets the date and time with time zone. It ignores milliseconds.
     * 
     * @param aYear the year.
     * @param aMonth the month of the year. January is 1, February is 2 etc.
     * @param aDate the day of month.
     * @param aHour the hour of the day in 24 format.
     * @param aMinute the minutes.
     * @param aSecond the seconds.
     * @param aTimeZone the time zone of the time provided.
     */
    public static void setDateTime(int aYear, int aMonth, int aDate, int aHour, int aMinute, int aSecond, TimeZone aTimeZone) {
        year.set(aYear);
        month.set(aMonth - 1);
        date.set(aDate);
        hour.set(aHour);
        minute.set(aMinute);
        second.set(aSecond);
        timezone.set(aTimeZone);
        setupCurrentTime();
    }

    /**
     * Sets the date and time.
     * 
     * @param aYear the year.
     * @param aMonth the month of the year. January is 1, February is 2 etc.
     * @param aDate the day of month.
     * @param aHour the hour of the day in 24 format.
     * @param aMinute the minutes.
     * @param aSecond the seconds.
     * @param aMillisecond the milliseconds.
     * @param aTimeZone the time zone of the time provided.
     */
    public static void setDateTime(int aYear, int aMonth, int aDate, int aHour, int aMinute, int aSecond, int aMillisecond, TimeZone aTimeZone) {
        year.set(aYear);
        month.set(aMonth - 1);
        date.set(aDate);
        hour.set(aHour);
        minute.set(aMinute);
        second.set(aSecond);
        millisecond.set(aMillisecond);
        timezone.set(aTimeZone);
        setupCurrentTime();
    }

    /**
     * Sets the date. Time is ignored.
     * 
     * @param aYear the year.
     * @param aMonth the month of the year. January is 1, February is 2 etc.
     * @param aDate the day of month.
     */
    public static void setDate(int aYear, int aMonth, int aDate) {
        year.set(aYear);
        month.set(aMonth - 1);
        date.set(aDate);
        setupCurrentTime();
    }
    
    /**
     * Sets the time. The date and time zone are ignored.
     * 
     * @param aHour the hour if the day in 24 hours format.
     * @param aMinute the minutes.
     * @param aSecond the seconds.
     * @param aMillisecond the milliseconds.
     */
    public static void setTime(int aHour, int aMinute, int aSecond, int aMillisecond) {
        hour.set(aHour);
        minute.set(aMinute);
        second.set(aSecond);
        millisecond.set(aMillisecond);
        setupCurrentTime();
    }
    
    /**
     * Sets the time. The date is ignored.
     * @param aHour the hour of the day in 24 hours format.
     * @param aMinute the minutes.
     * @param aSecond the seconds.
     * @param aMillisecond the milliseconds.
     * @param aTimeZone the time zone.
     */
    public static void setTime(int aHour, int aMinute, int aSecond, int aMillisecond, TimeZone aTimeZone) {
        hour.set(aHour);
        minute.set(aMinute);
        second.set(aSecond);
        millisecond.set(aMillisecond);
        timezone.set(aTimeZone);
        setupCurrentTime();
    }

    /**
     * Sets the time. The date, time zone and milliseconds are ignored.
     * @param aHour the hour of the day in 24 hours format.
     * @param aMinute the minutes.
     * @param aSecond the seconds.
     */
    public static void setTime(int aHour, int aMinute, int aSecond) {
        hour.set(aHour);
        minute.set(aMinute);
        second.set(aSecond);
        setupCurrentTime();
    }

    /**
     * Sets the time. The date, milliseconds are ignored.
     * @param aHour the hour of the day in 24 hours format.
     * @param aMinute the minutes.
     * @param aSecond the seconds.
     * @param aTimeZone the time zone.
     */
    public static void setTime(int aHour, int aMinute, int aSecond, TimeZone aTimeZone) {
        hour.set(aHour);
        minute.set(aMinute);
        second.set(aSecond);
        timezone.set(aTimeZone);
        setupCurrentTime();
    }

    /**
     * Sets the year. The rest of fields are ignored.
     * 
     * @param aYear the year.
     */
    public static void setYear(int aYear) {
        year.set(aYear);
        setupCurrentTime();
    }

    /**
     * Sets the month. The rest of fields are ignored.
     * 
     * @param aMonth the month in the year. January is 1, February is 2 etc.
     */
    public static void setMonth(int aMonth) {
        month.set(aMonth - 1);
        setupCurrentTime();
    }

    /**
     * Sets the day in the month. The rest of fields are ignored.
     * 
     * @param aDate the day in the month.
     */
    public static void setDate(int aDate) {
        date.set(aDate);
        setupCurrentTime();
    }

    /**
     * Sets the hour of the day. It is used 24 hours format. The rest of fields are ignored.
     * 
     * @param aHour the hour.
     */
    public static void setHour(int aHour) {
        hour.set(aHour);
        setupCurrentTime();
    }

    /**
     * Sets the minutes. The rest of fields are ignored.
     * 
     * @param aMinute the minutes.
     */
    public static void setMinute(int aMinute) {
        minute.set(aMinute);
        setupCurrentTime();
    }

    /**
     * Sets the seconds. The rest of fields are ignored.
     * 
     * @param aSecond the seconds.
     */
    public static void setSecond(int aSecond) {
        second.set(aSecond);
        setupCurrentTime();
    }

    /**
     * Sets the milliseconds. The rest of fields are ignored.
     * 
     * @param aMillisecond the milliseconds.
     */
    public static void setMillisecond(int aMillisecond) {
        millisecond.set(aMillisecond);
        setupCurrentTime();
    }

    /** Sets the time zone. The rest of fields are ignored.
     * 
     * @param aTimeZone the time zone.
     */
    public static void setTimezone(TimeZone aTimeZone) {
        timezone.set(aTimeZone);
        setupCurrentTime();
    }
    
    private static void setupCurrentTime() {
        Calendar calendar = new GregorianCalendar();
        
        if (year.get() == null) {
            year.set(calendar.get(Calendar.YEAR));
        }
        if (month.get() == null) {
            month.set(calendar.get(Calendar.MONTH));
        }
        if (date.get() == null) {
            date.set(calendar.get(Calendar.DATE));
        }
        if (hour.get() == null) {
            hour.set(calendar.get(Calendar.HOUR_OF_DAY));
        }
        if (minute.get() == null) {
            minute.set(calendar.get(Calendar.MINUTE));
        }
        if (second.get() == null) {
            second.set(calendar.get(Calendar.SECOND));
        }
        if (millisecond.get() == null) {
            millisecond.set(calendar.get(Calendar.MILLISECOND));
        }
        if (timezone.get() == null) {
            timezone.set(TimeZone.getDefault());
        }
        setupTime.set(System.currentTimeMillis());
    }
    
}
