/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono.runners;

import com.qitsoft.qimono.internal.QimonoClassLoader;
import com.qitsoft.qimono.internal.modifiers.CalendarInterceptModifier;
import com.qitsoft.qimono.internal.modifiers.DateInterceptModifier;
import com.qitsoft.qimono.internal.modifiers.FinalMethodsModifier;
import com.qitsoft.qimono.internal.modifiers.NonPublicFieldsModifier;
import com.qitsoft.qimono.internal.modifiers.NonPublicMethodsModifier;
import com.qitsoft.qimono.internal.modifiers.StaticMethodsModifier;
import com.qitsoft.qimono.internal.modifiers.SystemTimeInterceptModifier;
import com.qitsoft.qimono.internal.substitutors.DateSubstitutor;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.reflect.Constructor;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.junit.runner.Description;
import org.junit.runner.Runner;
import org.junit.runner.notification.RunListener;
import org.junit.runner.notification.RunNotifier;
import org.junit.runners.BlockJUnit4ClassRunner;
import org.junit.runners.model.InitializationError;

/**
 * The JUnit runner used for tests which requires Qimono features.
 * In the header of those tests you should use {@code @RunWith(QimonoJUnitRunner.class)}
 * to be able to use Qimono features.
 * This runner allows using parent runner. The QimonoJUnitRunner does not executes itself the tests it delegates the 
 * testing purposes to the parent runner. It only sets the {@link QimonoClassLoader}
 * to instrument the classes. By default is used {@link org.junit.runners.BlockJUnit4ClassRunner} runner.
 * To use another runner use {@link ParentRunner} annotation.
 * 
 * Also you can use the {@link Filter} annotation to set the classes and packages
 * ignored for instrumenting.
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public class QimonoJUnitRunner extends Runner {

    /**
     * The annotation used to set the parent runner if it is another than
     * {@link org.junit.runners.BlockJUnit4ClassRunner}. You are unable to use
     * {@link org.junit.runners.Suite} because it is a special runner for 
     * suite tests. Instead you can use the Qimono runner for each of enclosed tests.
     * 
     */
    @Retention(RetentionPolicy.RUNTIME)
    @Target(ElementType.TYPE)
    public static @interface ParentRunner {
        /**
         * The annotation value for parent runner class. By default is
         * {@link org.junit.runners.BlockJUnit4ClassRunner}.
         */
        Class<? extends Runner> value() default BlockJUnit4ClassRunner.class;
    }

    /**
     * The annotation used to declare the classes which should not be instrumented.
     * This annotation allows resolve conflicts between classes loaded by 
     * {@link QimonoClassLoader} and the classes loaded outside of it.
     * 
     * For this annotation you should set the full names of the classes or the 
     * names of packages. The packages should be written using ".*" at the end of 
     * the name.
     * 
     * <pre><code>
     * @Filter(value={"com.qitsoft.package.*","com.qitsoft.TheClass"})
     * // It is filtered the package com.qitsoft.package
     * // and the class com.qitsoft.TheClass
     * </code></pre>
     */
    @Retention(RetentionPolicy.RUNTIME)
    @Target(ElementType.TYPE)
    public static @interface Filter {
        /**
         * The list of full names of classes or packages (.* at the end).
         * @return 
         */
        String[] value();
    }

    /**
     * The constant declaring the default parent runner. It is set to {@link org.junit.runners.BlockJUnit4ClassRunner}.
     */
    protected static final Class<BlockJUnit4ClassRunner> DEFAULT_PARENT_RUNNER = BlockJUnit4ClassRunner.class;

    private Class testClass;
    private Runner parentRunner;
    private QimonoClassLoader classLoader;
    
    /**
     * The constructor used to instantiate the runner and to pass to it the test class
     * to run with.
     * 
     * @param testClass the test class to run.
     * @throws InitializationError  is some initialization error appears.
     */
    public QimonoJUnitRunner(Class testClass) throws InitializationError {
        this.classLoader = createClassloader(testClass);

        try {
            this.testClass = classLoader.loadClass(testClass.getName());
        } catch (ClassNotFoundException ex) {
            throw new InitializationError(ex);
        }

        parentRunner = createParentRunner(getParentRunnerClass());
    }

    /**
     * Initializes the filters set by {@link Filter} annotation.
     * 
     * @param classloader the class loader used to load the test class and other classes used by tests.
     * @param testClass the test class.
     */
    private void initializeFilters(QimonoClassLoader classloader, Class testClass) {
        for(String filteredObject : getFilteredObjects(testClass)) {
            if (filteredObject.endsWith(".*")) {
                classloader.addExcludePackage(filteredObject.replaceFirst("\\.\\*", ""));
            } else {
                classloader.addExcludeClass(filteredObject);
            }
        }
    }

    /**
     * Creates the class loader. This method also initializes the class loader with modifiers.
     * 
     * @param testClass the test class for which to create the class loader.
     * @return the class loader.
     */
    private QimonoClassLoader createClassloader(Class testClass) {
        QimonoClassLoader cl = new QimonoClassLoader();
        cl.addIncludeClass(testClass);
        initializeFilters(cl, testClass);

        cl.addClassModifier(new NonPublicMethodsModifier());
        cl.addClassModifier(new NonPublicFieldsModifier());
        cl.addClassModifier(new FinalMethodsModifier());
        StaticMethodsModifier.configureClassloader(cl);
// TODO: Should be implemented in futher versions
//        ConstructorsModifier.configureClassloader(cl);
        DateInterceptModifier.configureClassloader(cl);
        CalendarInterceptModifier.configureClassloader(cl);
        SystemTimeInterceptModifier.configureClassloader(cl);

        Thread.currentThread().setContextClassLoader(cl);

        return cl;
    }

    /**
     * Creates the parent runner. 
     * 
     * @param parentRunnerClass the class of the parent runner. 
     * @return the instance of the parent runner class.
     * @throws InitializationError if the parent runner class is null or another exception
     *      during initialization appeared.
     */
    private Runner createParentRunner(Class<? extends Runner> parentRunnerClass) throws InitializationError {
        if (parentRunnerClass == null) {
            throw new InitializationError("Cannot find the runner");
        }

        Constructor<? extends Runner> constructorMethod;
        try {
            constructorMethod = parentRunnerClass.getConstructor(Class.class);
            Object result = constructorMethod.newInstance(testClass);
            return (Runner) result;
        } catch (Exception ex) {
            try {
                Logger.getLogger(QimonoJUnitRunner.class.getName()).log(Level.WARNING, ex.getMessage(), ex);
                constructorMethod = parentRunnerClass.getConstructor();
                return constructorMethod.newInstance();
            } catch (Exception e) {
                throw new InitializationError(e);
            }
        }
    }

    /**
     * Returns the parent runner class. If the {@link ParentRunner} annotation is used
     * it returns that class otherwise will return the default parent runner.
     * @return the parent runner used to run the tests.
     * @throws InitializationError if appears some errors during initialization.
     */
    private Class<? extends Runner> getParentRunnerClass() throws InitializationError {
        ParentRunner parentRunnerAnnotation = (ParentRunner) testClass.getAnnotation(ParentRunner.class);
        Class<? extends Runner> parentRunnerClass;
        if (parentRunnerAnnotation != null && parentRunnerAnnotation.value() != null) {
            parentRunnerClass = parentRunnerAnnotation.value();
        } else {
            parentRunnerClass = DEFAULT_PARENT_RUNNER;
        }

        return parentRunnerClass;
    }

    /**
     * Returns the filtered objects declared by the {@link Filter} annotation.
     * If there is no annotation declared will return an empty array.
     * 
     * @param testClass the test class for which the annotation is declared.
     * @return the list of names passed into annotation.
     */
    private String[] getFilteredObjects(Class testClass) {
        Filter filterAnnotation = (Filter) testClass.getAnnotation(Filter.class);
        if (filterAnnotation != null && filterAnnotation.value() != null) {
            return filterAnnotation.value();
        } else {
            return new String[0];
        }
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public Description getDescription() {
        return parentRunner.getDescription();
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public void run(RunNotifier notifier) {
        notifier.addListener(new RunListener() {

            @Override
            public void testStarted(Description description) throws Exception {
                classLoader.resetSyntheticMocks();
                DateSubstitutor.reset();
                super.testStarted(description);
            }
            
        });
        parentRunner.run(notifier);
    }

}
