/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono.stubbing;

import java.lang.reflect.Field;
import org.mockito.exceptions.misusing.UnfinishedStubbingException;

/**
 * The class used to declare stubbing of the fields. 
 * This class allows setting or retrieving the values of private, protected, package (default)
 * fields.
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public class FieldStubbing<T> {

    private Object object;
    private String fieldName;

    /**
     * The constructor used to create the stubbing object.
     * 
     * @param object the object declared the field.
     * @param fieldName the name of the field.
     */
    public FieldStubbing(Object object, String fieldName) {
        this.object = object;
        this.fieldName = fieldName;
    }

    /**
     * The constructor used to create the stubbing object.
     * 
     * @param fieldName the field name to work with.
     */
    public FieldStubbing(String fieldName) {
        this.fieldName = fieldName;
    }

    /**
     * Declares the object owned the field.
     * 
     * @param object the object 
     * @return this instance for chain.
     */
    public FieldStubbing<T> from(Object object) {
        this.object = object;
        return this;
    }

    /**
     * Sets the value of the field. If the object and field name wasn't
     * declared before it will throw {@link NoSuchFieldException}. If you are trying 
     * to set the value to the field of incompatible type will throw {@link IllegalArgumentException}.
     * 
     * @param value the value to set to the field.
     * @throws NoSuchFieldException if the stubbing wasn't completely initialized .
     * @throws IllegalArgumentException is the value you are trying to set is invalid for that field.
     */
    public void set(T value) throws NoSuchFieldException, IllegalArgumentException {
        try {
            getField().set(object, value);
        } catch (IllegalAccessException ex) {
            throw new RuntimeException(ex);
        }
    }

    /**
     * Returns the value of the field.
     * 
     * @return the value of the field declared before.
     * @throws NoSuchFieldException if the object or field name wasn't declared 
     *      on the steps before or such field is not of that object.
     */
    public T get() throws NoSuchFieldException {
        try {
            return (T) getField().get(object);
        } catch (IllegalArgumentException ex) {
            throw new RuntimeException(ex);
        } catch (IllegalAccessException ex) {
            throw new RuntimeException(ex);
        }
    }

    /**
     * Returns the {@link java.lang.reflect.Field} object using the data collected 
     * during initialization of the object.
     * 
     * @return the {@link java.lang.reflect.Field} object of the field.
     * @throws UnfinishedStubbingException if the stubbing wasn't finished.
     * @throws NoSuchFieldException if such field is not declared in that object.
     * @throws IllegalArgumentException if an error appears during obtaining the field.
     * @throws SecurityException security exception during obtaining the field.
     */
    protected Field getField() throws UnfinishedStubbingException, NoSuchFieldException, IllegalArgumentException, SecurityException {
        if (object == null) {
            throw new UnfinishedStubbingException("You should use QiMockito.field() or FieldWrapper.from() methods to setup the stubbing object.");
        }
        if (fieldName == null) {
            throw new UnfinishedStubbingException("You should use QiMockito.field() method to setup the stubbing method.");
        }

        return object.getClass().getField(fieldName);
    }

}
