/*
 * Copyright (C) 2011 QitSoft Inc.
 *
 * This file is part of Qimono.
 *
 * Qimono is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package com.qitsoft.qimono.stubbing;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import org.mockito.Mockito;
import org.mockito.exceptions.misusing.MissingMethodInvocationException;
import org.mockito.exceptions.misusing.UnfinishedStubbingException;
import org.mockito.stubbing.Answer;
import org.mockito.stubbing.OngoingStubbing;

/**
 * The stunning object used to stub using reflection. This class can be used
 * to stub the methods by name and argument types. This stubbing object could be
 * used to stub the private/protected/package methods and public methods too.
 * 
 * @author Serghei Soloviov <serj@qitsoft.com>
 */
public class ReflectiveOngoingStubbing<T> implements OngoingStubbing<T>{

    private Object object;
    private String methodName;
    private Class[] argTypes;
    private Object[] args;

    /**
     * Declares the method to stub by name and arguments of the specific object.
     * 
     * @param object the object whom the method belongs.
     * @param methodName the method name
     * @param argTypes the list of argument types of the method.
     * @return this object to chain calls.
     */
    public ReflectiveOngoingStubbing<T> call(Object object, String methodName, Class... argTypes) {
        this.object = object;
        this.methodName = methodName;
        this.argTypes = argTypes;
        return this;
    }

    /**
     * Declares the method to stub by name and argument types. The mock object
     * should be used using {@link #from(java.lang.Class) } or {@link #from(java.lang.Object) } 
     * call.
     * 
     * @param methodName the method name.
     * @param argTypes the list of argument types.
     * @return this object for chain calls.
     */
    public ReflectiveOngoingStubbing<T> call(String methodName, Class... argTypes) {
        this.methodName = methodName;
        this.argTypes = argTypes;
        return this;
    }

    /**
     * Declares the mock object whose method should be stubbed.
     * 
     * @param object the mock object.
     * @return this object for chained calls.
     */
    public ReflectiveOngoingStubbing<T> from(Object object) {
        this.object = object;
        return this;
    }

    /**
     * Declares the object whose method should be stubbed. This method is used
     * to stub static methods of this class.
     * 
     * @param clazz the class whose methods should be stubbed.
     * @return this object for chained calls.
     */
    public ReflectiveOngoingStubbing<T> from(Class clazz) {
        this.object = clazz;
        return this;
    }

    /**
     * Pass the arguments of the methods for which is performed stubbing.
     * 
     * @param args the argument values
     * @return this object for chaining purposes.
     */
    public ReflectiveOngoingStubbing<T> withArgs(Object... args) {
        this.args = args;
        return this;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public OngoingStubbing<T> thenReturn(T value) {
        return stubMethod().thenReturn(value);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public OngoingStubbing<T> thenReturn(T value, T... values) {
        return stubMethod().thenReturn(value, values);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public OngoingStubbing<T> thenThrow(Throwable... throwables) {
        return stubMethod().thenThrow(throwables);
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public OngoingStubbing<T> thenCallRealMethod() {
        return stubMethod().thenCallRealMethod();
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public OngoingStubbing<T> thenAnswer(Answer<?> answer) {
        return stubMethod().thenAnswer(answer);
    }

    /**
     * Returns the stubbed method. If the object wasn't initialized till the end 
     * the {@link org.mockito.exceptions.misusing.UnfinishedStubbingException}
     * is thrown.
     * 
     * @return the stubbing object of {@link OngoingStubbing} type.
     */
    private OngoingStubbing<T> stubMethod() {
        if (object == null) {
            throw new UnfinishedStubbingException("You should use call() or from() method to setup the stubbing object.");
        }
        if (methodName == null) {
            throw new UnfinishedStubbingException("You should use call() method to setup the stubbing method.");
        }

        try {
            Class clazz = null;
            if (object instanceof Class) {
                clazz = (Class) object;
            } else {
                clazz = object.getClass();
            }
            Method method = clazz.getMethod(methodName, argTypes);
            T result = (T) method.invoke(object, args);
            return Mockito.when(result);
        } catch (IllegalAccessException ex) {
            throw new RuntimeException(ex);
        } catch (IllegalArgumentException ex) {
            throw new RuntimeException(ex);
        } catch (InvocationTargetException ex) {
            throw new RuntimeException(ex);
        } catch (NoSuchMethodException ex) {
            throw new MissingMethodInvocationException(ex.getMessage());
        } catch (SecurityException ex) {
            throw new RuntimeException(ex);
        }

    }

}
